"""Test aggregation methods."""
import numpy as np
from numpy.testing import TestCase, assert_equal
from scipy import sparse

from pyamg.gallery import poisson, load_example
from pyamg.strength import (symmetric_strength_of_connection,
                            classical_strength_of_connection)
from pyamg.aggregation.aggregate import (standard_aggregation, naive_aggregation,
                                         pairwise_aggregation)

from collections import OrderedDict


class TestAggregate(TestCase):
    def setUp(self):
        self.cases = []

        # random matrices
        np.random.seed(2006482792)
        for N in [2, 3, 5]:
            self.cases.append(sparse.csr_matrix(np.random.rand(N, N)))

        # Poisson problems in 1D and 2D
        for N in [2, 3, 5, 7, 10, 11, 19]:
            self.cases.append(poisson((N,), format='csr'))
        for N in [2, 3, 5, 7, 8]:
            self.cases.append(poisson((N, N), format='csr'))

        for name in ['knot', 'airfoil', 'bar']:
            ex = load_example(name)
            self.cases.append(ex['A'].tocsr())

    def test_standard_aggregation(self):
        for A in self.cases:
            S = symmetric_strength_of_connection(A)

            (expected, expected_Cpts) = reference_standard_aggregation(S)
            (result, Cpts) = standard_aggregation(S)

            assert_equal((result - expected).nnz, 0)
            assert_equal(Cpts.shape[0], expected_Cpts.shape[0])
            assert_equal(np.setdiff1d(Cpts, expected_Cpts).shape[0], 0)

        # S is diagonal - no dofs aggregated
        S = sparse.spdiags([[1, 1, 1, 1]], [0], 4, 4, format='csr')
        (result, Cpts) = standard_aggregation(S)
        expected = np.array([[0], [0], [0], [0]])
        assert_equal(result.toarray(), expected)
        assert_equal(Cpts.shape[0], 0)

    def test_naive_aggregation(self):
        for A in self.cases:
            S = symmetric_strength_of_connection(A)

            (expected, expected_Cpts) = reference_naive_aggregation(S)
            (result, Cpts) = naive_aggregation(S)

            assert_equal((result - expected).nnz, 0)
            assert_equal(Cpts.shape[0], expected_Cpts.shape[0])
            assert_equal(np.setdiff1d(Cpts, expected_Cpts).shape[0], 0)

        # S is diagonal - no dofs aggregated
        S = sparse.spdiags([[1, 1, 1, 1]], [0], 4, 4, format='csr')
        (result, Cpts) = naive_aggregation(S)
        expected = np.eye(4)
        assert_equal(result.toarray(), expected)
        assert_equal(Cpts.shape[0], 4)

    def test_pairwise_aggregation(self):
        for A in self.cases:
            S = classical_strength_of_connection(A, theta=0.25, norm='min')
            (expected, expected_Cpts) = reference_pairwise_aggregation(S)
            (result, Cpts) = pairwise_aggregation(A, matchings=1, theta=0.25, norm='min')

            assert_equal((result - expected).nnz, 0)
            assert_equal(Cpts.shape[0], expected_Cpts.shape[0])
            assert_equal(np.setdiff1d(Cpts, expected_Cpts).shape[0], 0)

        # S is diagonal - no dofs aggregated
        S = sparse.spdiags([[1.0, 1.0, 1.0, 1.0]], [0], 4, 4, format='csr')
        (result, Cpts) = pairwise_aggregation(S, matchings=1, theta=0.25, norm='min')
        expected = np.eye(4)
        assert_equal(result.todense(), expected)
        assert_equal(Cpts.shape[0], 4)


class TestComplexAggregate(TestCase):
    def setUp(self):
        self.cases = []

        # Poisson problems in 2D
        for N in [2, 3, 5, 7, 8]:
            A = poisson((N, N), format='csr')
            A.data = A.data + 0.001j*np.random.rand(A.nnz)
            self.cases.append(A)

    def test_standard_aggregation(self):
        for A in self.cases:
            S = symmetric_strength_of_connection(A)

            (expected, expected_Cpts) = reference_standard_aggregation(S)
            (result, Cpts) = standard_aggregation(S)

            assert_equal((result - expected).nnz, 0)
            assert_equal(Cpts.shape[0], expected_Cpts.shape[0])
            assert_equal(np.setdiff1d(Cpts, expected_Cpts).shape[0], 0)

    def test_naive_aggregation(self):
        for A in self.cases:
            S = symmetric_strength_of_connection(A)

            (expected, expected_Cpts) = reference_naive_aggregation(S)
            (result, Cpts) = naive_aggregation(S)

            assert_equal((result - expected).nnz, 0)
            assert_equal(Cpts.shape[0], expected_Cpts.shape[0])
            assert_equal(np.setdiff1d(Cpts, expected_Cpts).shape[0], 0)


# reference implementations for unittests  #
# note that this method only tests the current implementation, not
# all possible implementations
def reference_standard_aggregation(C):
    S = np.array_split(C.indices, C.indptr[1:-1])

    n = C.shape[0]

    R = set(range(n))
    j = 0
    Cpts = []

    aggregates = np.empty(n, dtype=C.indices.dtype)
    aggregates[:] = -1

    # Pass #1
    for i, row in enumerate(S):
        Ni = set(row) | {i}

        if Ni.issubset(R):
            Cpts.append(i)
            R -= Ni
            for x in Ni:
                aggregates[x] = j
            j += 1

    # Pass #2
    Old_R = R.copy()
    for i, row in enumerate(S):
        if i not in R:
            continue

        for x in row:
            if x not in Old_R:
                aggregates[i] = aggregates[x]
                R.remove(i)
                break

    # Pass #3
    for i, row in enumerate(S):
        if i not in R:
            continue
        Ni = set(row) | {i}
        Cpts.append(i)

        for x in Ni:
            if x in R:
                aggregates[x] = j
            j += 1

    assert len(R) == 0

    Pj = aggregates
    Pp = np.arange(n+1)
    Px = np.ones(n)

    return sparse.csr_matrix((Px, Pj, Pp)), np.array(Cpts)


def reference_naive_aggregation(C):
    S = np.array_split(C.indices, C.indptr[1:-1])
    n = C.shape[0]
    aggregates = np.empty(n, dtype=C.indices.dtype)
    aggregates[:] = -1  # aggregates[j] denotes the aggregate j is in
    R = np.zeros((0,))     # R stores already aggregated nodes
    j = 0               # j is the aggregate counter
    Cpts = []

    # Only one aggregation pass
    for i, row in enumerate(S):

        # if i isn't already aggregated, grab all his neighbors
        if aggregates[i] == -1:
            unaggregated_neighbors = np.setdiff1d(row, R)
            aggregates[unaggregated_neighbors] = j
            aggregates[i] = j
            j += 1
            R = np.union1d(R, unaggregated_neighbors)
            R = np.union1d(R, np.array([i]))
            Cpts.append(i)
        else:
            pass

    assert np.unique(R).shape[0] == n

    Pj = aggregates
    Pp = np.arange(n+1)
    Px = np.ones(n)

    return sparse.csr_matrix((Px, Pj, Pp)), np.array(Cpts)


def reference_pairwise_aggregation(C):
    S = np.array_split(C.indices, C.indptr[1:-1])
    data = np.array_split(C.data, C.indptr[1:-1])
    n = C.shape[0]
    aggregates = np.empty(n, dtype=C.indices.dtype)
    aggregates[:] = -1     # aggregates[j] denotes the aggregate j is in
    R = np.zeros((0,))     # R stores already aggregated nodes
    aggregate_count = 0               # aggregate_count is the aggregate counter
    Cpts = []
    m = np.zeros(n, dtype=C.indices.dtype)
    for row in S:
        m[row] = m[row] + 1

    max_m = max(m)

    # using a ordereddict here as there's no orderedset in python
    mmap = [OrderedDict() for _ in range(0, max_m+1)]

    for i in range(n):
        mmap[m[i]][i] = True

    count = 0
    aggregate_count = 0
    while (count < n):
        for k in range(0, max_m+1):
            if mmap[k]:
                i = list(mmap[k].keys())[0]
                break

        row = S[i]
        R = np.union1d(R, np.array([i]))
        aggregate_set = [i]
        aggregates[i] = aggregate_count
        max_aij = -np.inf
        max_aij_index = -1

        for k, j in enumerate(row):
            if j not in R and data[i][k] >= max_aij:
                max_aij = data[i][k]
                max_aij_index = j

        if max_aij_index != -1:
            j = max_aij_index
            aggregate_set.append(j)
            R = np.union1d(R, np.array([j]))
            aggregates[j] = aggregate_count

        Cpts.append(i)

        # Remove the aggregated nodes from mmap
        for j in aggregate_set:
            del mmap[m[j]][j]

        # Reduce m of the neighbors of the aggregated nodes
        for j in aggregate_set:
            row = S[j]
            unaggregated_neighbors = np.setdiff1d(row, R)
            for neighbour in unaggregated_neighbors:
                del mmap[m[neighbour]][neighbour]
                m[neighbour] = m[neighbour] - 1
                mmap[m[neighbour]][neighbour] = True

        count += len(aggregate_set)
        aggregate_count += 1

    assert (np.unique(R).shape[0] == n)

    Pj = aggregates
    Pp = np.arange(n+1)
    Px = np.ones(n)

    return sparse.csr_matrix((Px, Pj, Pp)), np.array(Cpts)



"""Generate matrix and right-hand side for upwind FD discretization of 2D advection."""

import numpy as np
from .stencil import stencil_grid


def advection_2d(grid, theta=np.pi/4.0, l_bdry=1.0, b_bdry=1.0):
    """
    Generate matrix and RHS for upwind FD discretization of 2D advection.

        (cos(theta),sin(theta)) dot grad(u) = 0,

    with inflow boundaries on the left and bottom of the domain. Assume uniform
    grid spacing, dx=dy, even for grid[0] != grid[1].

    Parameters
    ----------
    grid  : tuple
        (ny, nx) number of points in y and x
    theta : float, optional
        Rotation angle `theta` in radians defines direction of advection
        (cos(theta),sin(theta))
    l_bdry : float, array
        left boundary value. If float, then constant in-flow boundary value
        applied. If array, then length of array must be equal to ny = grid[0],
        and this array defines non-constant boundary value on the left.
    b_bdry : float, array
        bottom boundary value. If float, then constant in-flow boundary value
        applied. If array, then length of array must be equal to nx = grid[1],
        and this array defines non-constant boundary value on the bottom.

    Returns
    -------
    A : csr_matrix
        Defines 2D FD upwind discretization, with boundary
    rhs : array
        Defines right-hand-side with boundary contributions

    See Also
    --------
    poisson

    Examples
    --------
    >>> from numpy import pi
    >>> from pyamg.gallery import advection_2d
    >>> A, rhs = advection_2d( (4,4), theta=pi/4)
    >>> print(A.toarray().round(4))
        [[ 1.4142  0.      0.     -0.7071  0.      0.      0.      0.      0.    ]
         [-0.7071  1.4142  0.      0.     -0.7071  0.      0.      0.      0.    ]
         [ 0.     -0.7071  1.4142  0.      0.     -0.7071  0.      0.      0.    ]
         [ 0.      0.      0.      1.4142  0.      0.     -0.7071  0.      0.    ]
         [ 0.      0.      0.     -0.7071  1.4142  0.      0.     -0.7071  0.    ]
         [ 0.      0.      0.      0.     -0.7071  1.4142  0.      0.     -0.7071]
         [ 0.      0.      0.      0.      0.      0.      1.4142  0.      0.    ]
         [ 0.      0.      0.      0.      0.      0.     -0.7071  1.4142  0.    ]
         [ 0.      0.      0.      0.      0.      0.      0.     -0.7071  1.4142]]

    """
    grid = tuple(grid)
    if len(grid) != 2:
        raise ValueError('grid must be a length 2 tuple, '
                         'describe number of points in x and y')
    if theta <= 0 or theta >= np.pi/2:
        raise ValueError('theta must be in (0, pi/2)')

    # First-order upwind FD for dx and dy in (cos(theta),sin(theta)) \nabla u.
    w1 = np.cos(theta)
    w2 = np.sin(theta)
    st = np.array([[0, 0, 0], [-w1, w1+w2, 0], [0, -w2, 0]])
    A = stencil_grid(st, grid).tocsr()

    # Assume left and bottom of domain to be in-flow boundary
    # From
    #   grid=(ny,nx)
    #   np.arange(np.prod(grid)).reshape((grid))
    # We get boundary DOFs
    l_bdofs = np.array([i*grid[1] for i in range(0, grid[0])])
    b_bdofs = np.array([grid[1]*(grid[0]-1)+i for i in range(0, grid[1])])
    all_bdofs = np.concatenate((l_bdofs, b_bdofs))
    int_dofs = [i for i in range(0, A.shape[0]) if i not in all_bdofs]

    # Convert boundary values to array
    if np.isscalar(l_bdry):
        l_bdry = np.full(grid[0], l_bdry)
    elif l_bdry.shape[0] != grid[0]:
        raise ValueError('left boundary data does not match boundary size')

    if np.isscalar(b_bdry):
        b_bdry = np.full(grid[0], b_bdry)
    elif b_bdry.shape[0] != grid[1]:
        raise ValueError('bottom boundary data does not match boundary size')

    # Eliminate boundary DOFs
    bdry = np.concatenate((l_bdry.flatten(), b_bdry.flatten()))
    rhs = -A[int_dofs, :][:, all_bdofs]*bdry
    A = A[int_dofs, :][:, int_dofs]

    return A, rhs



"""Matrix gallery of model problems.

Functions
---------
    - poisson() : Poisson problem using Finite Differences
    - linear_elasticity() : Linear Elasticity using Finite Elements
    - stencil_grid() : General stencil generation from 1D, 2D, and 3D
    - diffusion_stencil_2d() : 2D rotated anisotropic FE/FD stencil
"""

from . import elasticity
from . import laplacian
from . import stencil
from . import diffusion

from .elasticity import linear_elasticity, linear_elasticity_p1
from .example import load_example
from .laplacian import poisson, gauge_laplacian
from .stencil import stencil_grid
from .mesh import regular_triangle_mesh
from .diffusion import diffusion_stencil_2d
from .advection import advection_2d

from .random_sparse import sprand
from .demo import demo

__all__ = [
    'elasticity', 'laplacian', 'stencil', 'diffusion',
    'linear_elasticity', 'linear_elasticity_p1',
    'load_example', 'advection_2d',
    'poisson', 'gauge_laplacian',
    'stencil_grid',
    'regular_triangle_mesh',
    'diffusion_stencil_2d',
    'sprand',
    'demo',
]



"""Approximate idealkrestriction AMG."""

from copy import deepcopy
import numpy as np
from scipy.sparse import isspmatrix_csr, isspmatrix_bsr

from ..multilevel import MultilevelSolver
from ..relaxation.smoothing import change_smoothers
from ..strength import (classical_strength_of_connection,
                        symmetric_strength_of_connection, evolution_strength_of_connection,
                        distance_strength_of_connection, algebraic_distance,
                        affinity_distance, energy_based_strength_of_connection)
from ..util.utils import filter_matrix_rows
from ..classical.interpolate import (direct_interpolation, classical_interpolation,
                                     injection_interpolation, one_point_interpolation,
                                     local_air)
from .split import RS, PMIS, PMISc, CLJP, CLJPc
from .cr import CR


def air_solver(A,
               strength=('classical', {'theta': 0.3, 'norm': 'min'}),
               CF=('RS', {'second_pass': True}),
               interpolation='one_point',
               restrict=('air', {'theta': 0.05, 'degree': 2}),
               presmoother=None,
               postsmoother=('fc_jacobi', {'omega': 1.0, 'iterations': 1,
                                           'withrho': False, 'f_iterations': 2,
                                           'c_iterations': 1}),
               filter_operator=None,
               max_levels=20, max_coarse=20,
               keep=False, **kwargs):
    """Create a multilevel solver using approximate ideal restriction (AIR) AMG.

    Parameters
    ----------
    A : csr_matrix
        Square (non)symmetric matrix in CSR format
    strength : ['symmetric', 'classical', 'evolution', 'distance',
                'algebraic_distance','affinity', 'energy_based', None]
        Method used to determine the strength of connection between unknowns
        of the linear system.  Method-specific parameters may be passed in
        using a tuple, e.g. strength=('symmetric',{'theta' : 0.25 }). If
        strength=None, all nonzero entries of the matrix are considered strong.
    CF : {string} : default 'RS' with second pass
        Method used for coarse grid selection (C/F splitting)
        Supported methods are RS, PMIS, PMISc, CLJP, CLJPc, and CR.
    interpolation : {string} : default 'one_point'
        Options include 'direct', 'classical', 'inject' and 'one-point'.
    restrict : {string} : default distance-2 AIR, with theta = 0.05.
        Option is 'air' for local approximate ideal restriction (lAIR),
        with inner options specifying degree, strength tolerance, etc.
    presmoother : {string or dict} : default None
        Method used for presmoothing at each level.  Method-specific parameters
        may be passed in using a tuple.
    postsmoother : {string or dict}
        Postsmoothing method with the same usage as presmoother.
        postsmoother=('fc_jacobi', ... ) with 2 F-sweeps, 1 C-sweep is default.
    filter_operator : (bool, tol) : default None
        Remove small entries in operators on each level if True. Entries are
        considered "small" if |a_ij| < tol |a_ii|.
    max_levels: {integer} : default 20
        Maximum number of levels to be used in the multilevel solver.
    max_coarse: {integer} : default 20
        Maximum number of variables permitted on the coarse grid.
    keep: {bool} : default False
        Flag to indicate keeping strength of connection matrix (C) in
        hierarchy.

    Returns
    -------
    ml : multilevel_solver
        Multigrid hierarchy of matrices and prolongation operators

    Examples
    --------
    >>> from pyamg.gallery import poisson
    >>> from pyamg import air_solver
    >>> A = poisson((10,),format='csr')
    >>> ml = air_solver(A,max_coarse=3)

    Notes
    -----
    "coarse_solver" is an optional argument and is the solver used at the
    coarsest grid.  The default is a pseudo-inverse.  Most simply,
    coarse_solver can be one of ['splu', 'lu', 'cholesky, 'pinv',
    'gauss_seidel', ... ].  Additionally, coarse_solver may be a tuple
    (fn, args), where fn is a string such as ['splu', 'lu', ...] or a callable
    function, and args is a dictionary of arguments to be passed to fn.
    See [2001TrOoSc]_ for additional details.


    References
    ----------
    [1] Manteuffel, T. A., Münzenmaier, S., Ruge, J., & Southworth, B. S.
    (2019). Nonsymmetric reduction-based algebraic multigrid. SIAM
    Journal on Scientific Computing, 41(5), S242-S268.

    [2] Manteuffel, T. A., Ruge, J., & Southworth, B. S. (2018).
    Nonsymmetric algebraic multigrid based on local approximate ideal
    restriction (lAIR). SIAM Journal on Scientific Computing, 40(6),
    A4105-A4130.

    See Also
    --------
    aggregation.smoothed_aggregation_solver, multilevel_solver,
    aggregation.rootnode_solver, ruge_stuben_solver

    """
    # preprocess A
    A = A.asfptype()
    if A.shape[0] != A.shape[1]:
        raise ValueError('expected square matrix')
    if np.iscomplexobj(A.data):
        raise ValueError('AIR solver not verified for complex matrices')

    levels = [MultilevelSolver.Level()]
    levels[-1].A = A

    while len(levels) < max_levels and levels[-1].A.shape[0] > max_coarse:
        bottom = extend_hierarchy(levels, strength, CF, interpolation, restrict,
                                  filter_operator, keep)
        if bottom:
            break

    ml = MultilevelSolver(levels, **kwargs)
    change_smoothers(ml, presmoother, postsmoother)
    return ml


def extend_hierarchy(levels, strength, CF, interpolation, restrict, filter_operator, keep):
    """Extend the multigrid hierarchy."""
    def unpack_arg(v):
        if isinstance(v, tuple):
            return v[0], v[1]
        return v, {}

    # Filter operator. Need to keep original matrix on finest level for
    # computing residuals
    if (filter_operator is not None) and (filter_operator[1] != 0):
        if len(levels) == 1:
            A = deepcopy(levels[-1].A)
        else:
            A = levels[-1].A
        filter_matrix_rows(A, filter_operator[1], diagonal=True, lump=filter_operator[0])
    else:
        A = levels[-1].A

    # Check if matrix was filtered to be diagonal --> coarsest grid
    if A.nnz == A.shape[0]:
        return 1

    # Compute the strength-of-connection matrix C, where larger
    # C[i,j] denote stronger couplings between i and j.
    fn, kwargs = unpack_arg(strength)
    if fn == 'symmetric':
        C = symmetric_strength_of_connection(A, **kwargs)
    elif fn == 'classical':
        C = classical_strength_of_connection(A, **kwargs)
    elif fn == 'distance':
        C = distance_strength_of_connection(A, **kwargs)
    elif fn in ('ode', 'evolution'):
        C = evolution_strength_of_connection(A, **kwargs)
    elif fn == 'energy_based':
        C = energy_based_strength_of_connection(A, **kwargs)
    elif fn == 'algebraic_distance':
        C = algebraic_distance(A, **kwargs)
    elif fn == 'affinity':
        C = affinity_distance(A, **kwargs)
    elif fn is None:
        C = A
    else:
        raise ValueError(f'Unrecognized strength of connection method: {fn}')

    # Generate the C/F splitting
    fn, kwargs = unpack_arg(CF)
    if fn == 'RS':
        splitting = RS(C, **kwargs)
    elif fn == 'PMIS':
        splitting = PMIS(C, **kwargs)
    elif fn == 'PMISc':
        splitting = PMISc(C, **kwargs)
    elif fn == 'CLJP':
        splitting = CLJP(C, **kwargs)
    elif fn == 'CLJPc':
        splitting = CLJPc(C, **kwargs)
    elif fn == 'CR':
        splitting = CR(C, **kwargs)
    else:
        raise ValueError(f'Unknown C/F splitting method {CF}')

    # Make sure all points were not declared as C- or F-points
    num_fpts = np.sum(splitting)
    if (num_fpts == len(splitting)) or (num_fpts == 0):
        return 1

    # Generate the interpolation matrix that maps from the coarse-grid to the
    # fine-grid
    fn, kwargs = unpack_arg(interpolation)
    if fn == 'classical':
        P = classical_interpolation(A, C, splitting, **kwargs)
    elif fn == 'direct':
        P = direct_interpolation(A, C, splitting, **kwargs)
    elif fn == 'one_point':
        P = one_point_interpolation(A, C, splitting, **kwargs)
    elif fn == 'inject':
        P = injection_interpolation(A, splitting, **kwargs)
    else:
        raise ValueError(f'Unknown interpolation method {fn}')

    # Build restriction operator
    fn, kwargs = unpack_arg(restrict)
    if fn == 'air':
        R = local_air(A, splitting, **kwargs)
    else:
        raise ValueError(f'Unknown restriction method {fn}')

    # Store relevant information for this level
    if keep:
        levels[-1].C = C              # strength of connection matrix

    levels[-1].splitting = splitting.astype(bool)  # C/F splitting
    levels[-1].P = P                               # prolongation operator
    levels[-1].R = R                               # restriction operator

    # RAP = R*(A*P)
    A = R * A * P

    # Make sure coarse-grid operator is in correct sparse format
    if (isspmatrix_csr(P) and (not isspmatrix_csr(A))):
        A = A.tocsr()
    elif (isspmatrix_bsr(P) and (not isspmatrix_bsr(A))):
        A = A.tobsr()

    levels.append(MultilevelSolver.Level())
    levels[-1].A = A
    return 0



import matplotlib
import numpy as np
from cycler import cycler


def set_figure(fontsize=9, width=251.0, heightratio=None, height=None):
    r"""
    Parameters
    ----------
    fontsize : float
        sets the intended fontsize

    width : float
        sets the intended width in pts

    Notes
    -----
    To set equal to the columnwidth of the article:

    In the tex file 'Column width: \the\columnwidth' will print this size
    alternatively, '\message{Column width: \the\columnwidth}' will print to the log

    \linewidth should be used in place of \columnwidth if the figure is used
    within special enviroments (e.g. minipage)

    https://matplotlib.org/stable/tutorials/introductory/customizing.html
    https://scipy-cookbook.readthedocs.io/items/Matplotlib_LaTeX_Examples.html
    https://tex.stackexchange.com/questions/16942/difference-between-textwidth-linewidth-and-hsize
    """
    fig_width_pt = width
    inches_per_pt = 1.0/72.27               # Convert pt to inch
    fig_width = fig_width_pt*inches_per_pt  # width in inches

    if heightratio is None:
        heightratio = (np.sqrt(5)-1.0)/2.0  # Aesthetic ratio
    if height is None:
        fig_height = fig_width*heightratio      # height in inches
    else:
        fig_height = height*inches_per_pt
    fig_size = [fig_width, fig_height]
    params = {'backend': 'pdf',
              'text.usetex': True,
              'text.latex.preamble': r"""
                                      \usepackage{lmodern}
                                      \usepackage[cm]{sfmath}
                                      \usepackage[T1]{fontenc}
                                      """,
              # fonts
              'font.family': 'sans-serif',
              # font sizes
              'axes.labelsize': fontsize,
              'font.size': fontsize,
              'axes.titlesize': fontsize,
              'legend.fontsize': fontsize,
              'xtick.labelsize': fontsize,
              'ytick.labelsize': fontsize,
              # figure size
              'figure.figsize': fig_size,
              'figure.constrained_layout.use': True,
              # legend
              'legend.frameon': False,
              # spines
              'axes.spines.top': True,
              'axes.spines.right': True,
              # saving
              'savefig.bbox': 'tight',
              'savefig.pad_inches': 1/72,
              # grid
              'grid.color': '0.7',
              'grid.linewidth': 0.2,
              # ticks
              'xtick.direction': 'in',
              'ytick.direction': 'in',
              }
    matplotlib.rcParams.update(params)



"""Test adaptive SA."""
import warnings
import numpy as np
from numpy.testing import TestCase
from scipy.sparse import SparseEfficiencyWarning

from pyamg.gallery import poisson, linear_elasticity
from pyamg.aggregation import smoothed_aggregation_solver
from pyamg.aggregation.adaptive import adaptive_sa_solver


class TestAdaptiveSA(TestCase):
    def setUp(self):
        np.random.seed(3485190434)

    def test_poisson(self):
        A = poisson((50, 50), format='csr')

        [asa, work] = adaptive_sa_solver(A, num_candidates=1)
        sa = smoothed_aggregation_solver(A, B=np.ones((A.shape[0], 1)))

        b = np.random.rand(A.shape[0])

        residuals0 = []
        residuals1 = []

        sol0 = asa.solve(b, maxiter=20, tol=1e-10, residuals=residuals0)
        sol1 = sa.solve(b, maxiter=20, tol=1e-10, residuals=residuals1)
        del sol0, sol1

        conv_asa = (residuals0[-1] / residuals0[0]) ** (1.0 / len(residuals0))
        conv_sa = (residuals1[-1] / residuals1[0]) ** (1.0 / len(residuals1))

        # print "ASA convergence (Poisson)",conv_asa
        # print "SA convergence (Poisson)",conv_sa
        assert conv_asa < 1.2 * conv_sa

    def test_elasticity(self):
        warnings.filterwarnings('ignore', category=UserWarning,
                                message='Having less target vectors')
        A, B = linear_elasticity((35, 35), format='bsr')

        smoother = ('gauss_seidel', {'sweep': 'symmetric', 'iterations': 2})
        [asa, work] = adaptive_sa_solver(A, num_candidates=3,
                                         improvement_iters=5,
                                         prepostsmoother=smoother)
        sa = smoothed_aggregation_solver(A, B=B)

        b = np.random.rand(A.shape[0])

        residuals0 = []
        residuals1 = []

        sol0 = asa.solve(b, maxiter=20, tol=1e-10, residuals=residuals0)
        sol1 = sa.solve(b, maxiter=20, tol=1e-10, residuals=residuals1)
        del sol0, sol1

        conv_asa = (residuals0[-1] / residuals0[0]) ** (1.0 / len(residuals0))
        conv_sa = (residuals1[-1] / residuals1[0]) ** (1.0 / len(residuals1))

        # print "ASA convergence (Elasticity) %1.2e" % (conv_asa)
        # print "SA convergence (Elasticity) %1.2e" % (conv_sa)
        assert conv_asa < 1.3 * conv_sa

    def test_matrix_formats(self):
        warnings.filterwarnings('ignore', category=SparseEfficiencyWarning)

        # Do dense, csr, bsr and csc versions of A all yield the same solver
        A = poisson((7, 7), format='csr')
        cases = [A.tobsr(blocksize=(1, 1))]
        cases.append(A.tocsc())
        cases.append(A.toarray())

        np.random.seed(111908910)
        sa_old = adaptive_sa_solver(A, initial_candidates=np.ones((49, 1)),
                                    max_coarse=10)[0]
        for AA in cases:
            np.random.seed(111908910)
            sa_new = adaptive_sa_solver(AA,
                                        initial_candidates=np.ones((49, 1)),
                                        max_coarse=10)[0]
            assert (abs(np.ravel(sa_old.levels[-1].A.toarray()
                    - sa_new.levels[-1].A.toarray())).max()) < 0.01
            sa_old = sa_new


class TestComplexAdaptiveSA(TestCase):
    def setUp(self):
        np.random.seed(1985581638)

    def test_poisson(self):
        cases = []

        # perturbed Laplacian
        A = poisson((50, 50), format='csr')
        Ai = A.copy()
        Ai.data = Ai.data + 1e-5j * np.random.rand(Ai.nnz)
        cases.append((Ai, 0.25))

        # imaginary Laplacian
        Ai = 1.0j * A
        cases.append((Ai, 0.25))

        # JBS:  Not sure if this is a valid test case
        # imaginary shift
        # Ai = A + 1.1j*sparse.eye(A.shape[0], A.shape[1])
        # cases.append((Ai,0.8))

        for A, rratio in cases:
            [asa, work] = adaptive_sa_solver(A, num_candidates=1, symmetry='symmetric')
            # sa = smoothed_aggregation_solver(A, B = np.ones((A.shape[0],1)) )

            b = np.zeros((A.shape[0],))
            x0 = np.random.rand(A.shape[0],) + 1.0j * np.random.rand(A.shape[0],)

            residuals0 = []

            sol0 = asa.solve(b, x0=x0, maxiter=20, tol=1e-10,
                             residuals=residuals0)
            del sol0

            conv_asa = \
                (residuals0[-1] / residuals0[0]) ** (1.0 / len(residuals0))

            assert conv_asa < rratio

# class TestAugmentCandidates(TestCase):
#    def setUp(self):
#        self.cases = []
#
# two candidates
#
# block candidates
# self.cases.append((
#   csr_matrix((np.ones(9),array([0,0,0,1,1,1,2,2,2]),arange(10)),
#   shape=(9,3)), vstack((array([1]*9 + [0]*9),arange(2*9))).T ))
#
#    def test_first_level(self):
#        cases = []
#
# tests where AggOp includes all DOFs
#        cases.append((
#           csr_matrix((np.ones(4),array([0,0,1,1]),arange(5)),
#           shape=(4,2)), vstack((np.ones(4),arange(4))).T ))
#        cases.append((
#           csr_matrix((np.ones(9),array([0,0,0,1,1,1,2,2,2]),arange(10)),
#           shape=(9,3)), vstack((np.ones(9),arange(9))).T ))
#        cases.append((
#           csr_matrix((np.ones(9),array([0,0,1,1,2,2,3,3,3]),arange(10)),
#           shape=(9,4)), vstack((np.ones(9),arange(9))).T ))
#
# tests where AggOp excludes some DOFs
#        cases.append((
#           csr_matrix((np.ones(4),array([0,0,1,1]),array([0,1,2,2,3,4])),
#           shape=(5,2)), vstack((np.ones(5),arange(5))).T ))
#
# overdetermined blocks
#        cases.append((
#           csr_matrix((np.ones(4),array([0,0,1,1]),array([0,1,2,2,3,4])),
#           shape=(5,2)), vstack((np.ones(5),arange(5),arange(5)**2)).T  ))
#        cases.append((
#           csr_matrix(
#               (np.ones(6),array([1,3,0,2,1,0]),array([0,0,1,2,2,3,4,5,5,6])),
#           shape=(9,4)), vstack((np.ones(9),arange(9),arange(9)**2)).T ))
#        cases.append((
#           csr_matrix(
#               (np.ones(6),array([1,3,0,2,1,0]),array([0,0,1,2,2,3,4,5,5,6])),
#           shape=(9,4)), vstack((np.ones(9),arange(9))).T ))
#
#        def mask_candidate(AggOp,candidates):
# mask out all DOFs that are not included in the aggregation
#            candidates[diff(AggOp.indptr) == 0,:] = 0
#
#        for AggOp,fine_candidates in cases:
#
#            mask_candidate(AggOp,fine_candidates)
#
#            for i in range(1,fine_candidates.shape[1]):
#                Q_expected,R_expected =
#                   fit_candidates(AggOp,fine_candidates[:, :i+1])
#
#                old_Q, old_R = fit_candidates(AggOp,fine_candidates[:,:i])
#
#                Q_result,R_result = augment_candidates(AggOp, old_Q, old_R,
#                   fine_candidates[:, [i]])
#
# compare against SA method (which is assumed to be correct)
#                assert_almost_equal(Q_expected.toarray(),Q_result.toarray())
#                assert_almost_equal(R_expected,R_result)
#
# each fine level candidate should be fit exactly
#                assert_almost_equal(fine_candidates[:,:i+1],Q_result*R_result)
#                assert_almost_equal(
#                   Q_result*(Q_result.T*fine_candidates[:, :i+1]),
#                   fine_candidates[:, :i+1])



"""Test warnings used."""
import warnings
from numpy.testing import TestCase


class TestWarn(TestCase):
    def test_f(self):
        warnings.filterwarnings('ignore', message='another warning')
        warnings.warn('another warning!', stacklevel=2)



"""Examples stored in files."""

import os
from glob import glob

from scipy.io import loadmat

base_dir = os.path.split(__file__)[0]
example_dir = os.path.join(base_dir, 'example_data')
example_files = glob(os.path.join(example_dir, '*.mat'))
example_names = sorted(os.path.split(name)[1][:-4] for name in example_files)


def load_example(name):
    """Load an example problem by name.

    Parameters
    ----------
    name : string (e.g. 'airfoil')
        Name of the example to load

    Notes
    -----
    Each example is stored in a dictionary with the following keys:
        - 'A'        : sparse matrix
        - 'B'        : near-nullspace candidates
        - 'vertices' : dense array of nodal coordinates
        - 'elements' : dense array of element indices

    Current example names are:%s

    Examples
    --------
    >>> from pyamg.gallery import load_example
    >>> ex = load_example('knot')

    """
    if name not in example_names:
        raise ValueError(f'No example with name {name}')

    return loadmat(os.path.join(example_dir, name + '.mat'), struct_as_record=True)


# insert the example names into the docstring
load_example.__doc__ %= ('\n' + ' ' * 8).join([''] + example_names)



"""amg_core - a C++ implementation of AMG-related routines."""

from .evolution_strength import (apply_absolute_distance_filter, apply_distance_filter,
                                 min_blocks, evolution_strength_helper,
                                 incomplete_mat_mult_csr)
from .graph import (maximal_independent_set_serial, maximal_independent_set_parallel,
                    vertex_coloring_mis, vertex_coloring_jones_plassmann,
                    vertex_coloring_LDF,
                    cluster_node_incidence, cluster_center,
                    bellman_ford, lloyd_cluster, lloyd_cluster_exact,
                    maximal_independent_set_k_parallel,
                    breadth_first_search, connected_components)

from .krylov import (apply_householders, householder_hornerscheme, apply_givens)
from .linalg import (pinv_array, csc_scale_columns, csc_scale_rows, filter_matrix_rows)
from .relaxation import (gauss_seidel, bsr_gauss_seidel, gauss_seidel_indexed,
                         jacobi, bsr_jacobi,
                         jacobi_ne, gauss_seidel_ne, gauss_seidel_nr,
                         block_jacobi, block_gauss_seidel,
                         extract_subblocks, overlapping_schwarz_csr,
                         jacobi_indexed, bsr_jacobi_indexed, block_jacobi_indexed)
from .ruge_stuben import (classical_strength_of_connection_abs,
                          classical_strength_of_connection_min,
                          maximum_row_value,
                          rs_cf_splitting, rs_cf_splitting_pass2,
                          cljp_naive_splitting,
                          rs_direct_interpolation_pass1, rs_direct_interpolation_pass2,
                          cr_helper,
                          rs_classical_interpolation_pass1,
                          rs_classical_interpolation_pass2,
                          remove_strong_FF_connections)
from .smoothed_aggregation import (symmetric_strength_of_connection, standard_aggregation,
                                   naive_aggregation, pairwise_aggregation,
                                   fit_candidates,
                                   satisfy_constraints_helper, calc_BtB,
                                   incomplete_mat_mult_bsr, truncate_rows_csr)

from .air import (one_point_interpolation, approx_ideal_restriction_pass1,
                  approx_ideal_restriction_pass2, block_approx_ideal_restriction_pass2)

__all__ = [
    'apply_absolute_distance_filter',
    'apply_distance_filter',
    'min_blocks',
    'evolution_strength_helper',
    'incomplete_mat_mult_csr',
    #
    'maximal_independent_set_serial',
    'maximal_independent_set_parallel',
    'vertex_coloring_mis',
    'vertex_coloring_jones_plassmann',
    'vertex_coloring_LDF',
    'cluster_node_incidence',
    'cluster_center',
    'bellman_ford',
    'lloyd_cluster',
    'lloyd_cluster_exact',
    'maximal_independent_set_k_parallel',
    'breadth_first_search',
    'connected_components',
    #
    'apply_householders',
    'householder_hornerscheme',
    'apply_givens',
    #
    'pinv_array',
    'csc_scale_columns',
    'csc_scale_rows',
    'filter_matrix_rows',
    #
    'gauss_seidel',
    'bsr_gauss_seidel',
    'jacobi',
    'bsr_jacobi',
    'gauss_seidel_indexed',
    'jacobi_ne',
    'gauss_seidel_ne',
    'gauss_seidel_nr',
    'block_jacobi',
    'block_gauss_seidel',
    'extract_subblocks',
    'overlapping_schwarz_csr',
    'jacobi_indexed',
    'bsr_jacobi_indexed',
    'block_jacobi_indexed',
    #
    'classical_strength_of_connection_abs',
    'classical_strength_of_connection_min',
    'maximum_row_value',
    'rs_cf_splitting',
    'rs_cf_splitting_pass2',
    'cljp_naive_splitting',
    'rs_direct_interpolation_pass1',
    'rs_direct_interpolation_pass2',
    'cr_helper',
    'rs_classical_interpolation_pass1',
    'rs_classical_interpolation_pass2',
    'remove_strong_FF_connections',
    #
    'symmetric_strength_of_connection',
    'standard_aggregation',
    'naive_aggregation',
    'pairwise_aggregation',
    'fit_candidates',
    'satisfy_constraints_helper',
    'calc_BtB',
    'incomplete_mat_mult_bsr',
    'truncate_rows_csr',
    #
    'one_point_interpolation',
    'approx_ideal_restriction_pass1',
    'approx_ideal_restriction_pass2',
    'block_approx_ideal_restriction_pass2'
    #
]



"""VTK output functions.

Create coarse grid views and write meshes/primitives to .vtu files.  Use the
XML VTK format for unstructured meshes (.vtu)

This will use the XML VTK format for unstructured meshes, .vtu

See here for a guide:  http://www.vtk.org/pdf/file-formats.pdf
"""

import xml.dom.minidom
import numpy as np


def write_vtu(V, cells, pdata=None, pvdata=None, cdata=None, cvdata=None,
              fname='output.vtk'):
    """Write a .vtu file in xml format.

    Parameters
    ----------
    fname : {string}
        file to be written, e.g. 'mymesh.vtu'
    V : {array}
        Ndof x 3 (if 2, then expanded by 0)
        list of (x,y,z) point coordinates
    cells : {dictionary}
        Dictionary of with the keys
    pdata : {array}
        Ndof x Nfields array of scalar values for the vertices
    pvdata : {array}
        Nfields*3 x Ndof array of vector values for the vertices
    cdata : {dictionary}
        scalar valued cell data
    cvdata : {dictionary}
        vector valued cell data

    Returns
    -------
    writes a .vtu file for use in Paraview

    Notes
    -----
    - Poly data not supported
    - Non-Poly data is stored in Numpy array: Ncell x vtk_cell_info
    - Each I1 must be >=3
    - pdata = Ndof x Nfields
    - pvdata = 3*Ndof x Nfields
    - cdata,cvdata = list of dictionaries in the form of cells


    =====  =================== ============= ===
    keys   type                n points      dim
    =====  =================== ============= ===
       1   VTK_VERTEX:         1 point        2d
       2   VTK_POLY_VERTEX:    n points       2d
       3   VTK_LINE:           2 points       2d
       4   VTK_POLY_LINE:      n+1 points     2d
       5   VTK_TRIANGLE:       3 points       2d
       6   VTK_TRIANGLE_STRIP: n+2 points     2d
       7   VTK_POLYGON:        n points       2d
       8   VTK_PIXEL:          4 points       2d
       9   VTK_QUAD:           4 points       2d
       10  VTK_TETRA:          4 points       3d
       11  VTK_VOXEL:          8 points       3d
       12  VTK_HEXAHEDRON:     8 points       3d
       13  VTK_WEDGE:          6 points       3d
       14  VTK_PYRAMID:        5 points       3d
    =====  =================== ============= ===

    Examples
    --------
    >>> from pyamg.vis import write_vtu
    >>> import numpy as np
    >>> V = np.array([[0.0,0.0],
    ...               [1.0,0.0],
    ...               [2.0,0.0],
    ...               [0.0,1.0],
    ...               [1.0,1.0],
    ...               [2.0,1.0],
    ...               [0.0,2.0],
    ...               [1.0,2.0],
    ...               [2.0,2.0],
    ...               [0.0,3.0],
    ...               [1.0,3.0],
    ...               [2.0,3.0]])
    >>> E2V = np.array([[0,4,3],
    ...                 [0,1,4],
    ...                 [1,5,4],
    ...                 [1,2,5],
    ...                 [3,7,6],
    ...                 [3,4,7],
    ...                 [4,8,7],
    ...                 [4,5,8],
    ...                 [6,10,9],
    ...                 [6,7,10],
    ...                 [7,11,10],
    ...                 [7,8,11]])
    >>> E2edge = np.array([[0,1]])
    >>> E2point = np.array([2,3,4,5])
    >>> cells = {5:E2V,3:E2edge,1:E2point}
    >>> pdata=np.ones((12,2))
    >>> pvdata=np.ones((12*3,2))
    >>> cdata={5:np.ones((12,2)),3:np.ones((1,2)),1:np.ones((4,2))}
    >>> cvdata={5:np.ones((3*12,2)),3:np.ones((3*1,2)), 1:np.ones((3*4,2))}
    >>> write_vtu(V=V, cells=cells, fname='test.vtu')

    See Also
    --------
    write_mesh

    """
    # number of indices per cell for each cell type
    vtk_cell_info = [-1, 1, None, 2, None, 3, None, None, 4, 4, 4, 8, 8, 6, 5]

    # check fname
    if not isinstance(fname, str):
        raise ValueError('fname is assumed to be a string')

    # check V
    # get dimension and verify that it's 3d data
    Ndof, dim = V.shape
    if dim == 2:
        # always use 3d coordinates (x,y) -> (x,y,0)
        V = np.hstack((V, np.zeros((Ndof, 1))))

    # check cells
    # keys must ve valid (integer and not "None" in vtk_cell_info)
    # Cell data can't be empty for a non empty key
    for key in cells:
        if ((not isinstance(key, int)) or (key not in list(range(1, 15)))):
            raise ValueError('cell array must have positive integer keys in [1,14]')
        if (vtk_cell_info[key] is None) and (cells[key] is not None):
            # Poly data
            raise NotImplementedError('Poly Data not implemented yet')
        if cells[key] is None:
            raise ValueError(f'Cell array cannot be empty for key {key}')
        if np.ndim(cells[key]) != 2:
            cells[key] = cells[key].reshape((cells[key].size, 1))
        if vtk_cell_info[key] != cells[key].shape[1]:
            raise ValueError(f'Cell array has {cells[key].shape[1]} columns. '
                             f'Expected {vtk_cell_info[key]}.')

    # check pdata
    # must be Ndof x n_pdata
    n_pdata = 0
    if pdata is not None:
        if np.ndim(pdata) > 1:
            n_pdata = pdata.shape[1]
        else:
            n_pdata = 1
            pdata = pdata.reshape((pdata.size, 1))
        if pdata.shape[0] != Ndof:
            raise ValueError(f'pdata array should be length {Ndof} (not {pdata.shape[0]})')

    # check pvdata
    # must be 3*Ndof x n_pvdata
    n_pvdata = 0
    if pvdata is not None:
        if np.ndim(pvdata) > 1:
            n_pvdata = pvdata.shape[1]
        else:
            n_pvdata = 1
            pvdata = pvdata.reshape((pvdata.size, 1))
        if pvdata.shape[0] != 3*Ndof:
            raise ValueError(f'pvdata array should be of size {3*Ndof} '
                             f'(or multiples) (it is now {pvdata.shape[0]}).')

    # check cdata
    # must be Ncells x n_cdata for each key
    n_cdata = 0
    if cdata is not None:
        for key in cells:   # all valid now
            if np.ndim(cdata[key]) > 1:
                if n_cdata == 0:
                    n_cdata = cdata[key].shape[1]
                elif n_cdata != cdata[key].shape[1]:
                    raise ValueError('cdata dimension problem')
            else:
                n_cdata = 1
                cdata[key] = cdata[key].reshape((cdata[key].size, 1))
            if cdata[key].shape[0] != cells[key].shape[0]:
                raise ValueError(f'Size mismatch with cdata {cdata[key].shape[0]} '
                                 f'and cells {cells[key].shape[0]}.')
            if cdata[key] is None:
                raise ValueError(f'cdata array cannot be empty for key {key}')

    # check cvdata
    # must be Ncells*3 x n_cdata for each key
    n_cvdata = 0
    if cvdata is not None:
        for key in cells:   # all valid now
            if np.ndim(cvdata[key]) > 1:
                if n_cvdata == 0:
                    n_cvdata = cvdata[key].shape[1]
                elif n_cvdata != cvdata[key].shape[1]:
                    raise ValueError('cvdata dimension problem')
            else:
                n_cvdata = 1
                cvdata[key] = cvdata[key].reshape((cvdata[key].size, 1))
            if cvdata[key].shape[0] != 3 * cells[key].shape[0]:
                raise ValueError('size mismatch with cvdata and cells')
            if cvdata[key] is None:
                raise ValueError(f'cvdata array cannot be empty for key {key}')

    Ncells = 0
    cell_ind = np.empty((0,), dtype=np.int32)
    # zero indexed
    cell_offset = np.empty((0,), dtype='uint8')  # np.zeros((Ncells,1),dtype=uint8)
    cell_type = np.empty((0,), dtype='uint8')    # np.zeros((Ncells,1),dtype=uint8)

    cdata_all = None
    cvdata_all = None
    for key in cells:
        # non-Poly data
        sz = cells[key].shape[0]
        offset = cells[key].shape[1]

        Ncells += sz
        uu = np.ones((sz,), dtype='uint8')
        cell_ind = np.hstack((cell_ind, cells[key].ravel()))
        cell_offset = np.hstack((cell_offset, offset * uu))
        cell_type = np.hstack((cell_type, key * uu))

        if cdata is not None:
            if cdata_all is None:
                cdata_all = cdata[key]
            else:
                cdata_all = np.vstack((cdata_all, cdata[key]))

        if cvdata is not None:
            if cvdata_all is None:
                cvdata_all = cvdata[key]
            else:
                cvdata_all = np.vstack((cvdata_all, cvdata[key]))

    # doc element
    doc = xml.dom.minidom.Document()

    # vtk element
    root = doc.createElementNS('VTK', 'VTKFile')
    d = {'type': 'UnstructuredGrid', 'version': '0.1',
         'byte_order': 'LittleEndian'}
    _set_attributes(d, root)

    # unstructured element
    grid = doc.createElementNS('VTK', 'UnstructuredGrid')

    # piece element
    piece = doc.createElementNS('VTK', 'Piece')
    d = {'NumberOfPoints': str(Ndof), 'NumberOfCells': str(Ncells)}
    _set_attributes(d, piece)

    # POINTS
    # points element
    points = doc.createElementNS('VTK', 'Points')
    # data element
    points_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type': 'Float32', 'Name': 'vertices', 'NumberOfComponents': '3',
         'format': 'ascii'}
    _set_attributes(d, points_data)
    # string for data element
    points_data_str = doc.createTextNode(_a2s(V))

    # CELLS
    # points element
    cells = doc.createElementNS('VTK', 'Cells')
    # data element
    cells_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type': 'Int32', 'Name': 'connectivity', 'format': 'ascii'}
    _set_attributes(d, cells_data)
    # string for data element
    cells_data_str = doc.createTextNode(_a2s(cell_ind))
    # offset data element
    cells_offset_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type': 'Int32', 'Name': 'offsets', 'format': 'ascii'}
    _set_attributes(d, cells_offset_data)
    # string for data element
    cells_offset_data_str = doc.createTextNode(_a2s(cell_offset.cumsum()))
    # offset data element
    cells_type_data = doc.createElementNS('VTK', 'DataArray')
    d = {'type': 'UInt8', 'Name': 'types', 'format': 'ascii'}
    _set_attributes(d, cells_type_data)
    # string for data element
    cells_type_data_str = doc.createTextNode(_a2s(cell_type))

    # POINT DATA
    pointdata = doc.createElementNS('VTK', 'PointData')
    # pdata
    pdata_obj = []
    pdata_str = []
    for i in range(0, n_pdata):
        pdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type': 'Float32', 'Name': f'pdata {i}',
             'NumberOfComponents': '1', 'format': 'ascii'}
        _set_attributes(d, pdata_obj[i])
        pdata_str.append(doc.createTextNode(_a2s(pdata[:, i])))
    # pvdata
    pvdata_obj = []
    pvdata_str = []
    for i in range(0, n_pvdata):
        pvdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type': 'Float32', 'Name': f'pvdata {i}',
             'NumberOfComponents': '3', 'format': 'ascii'}
        _set_attributes(d, pvdata_obj[i])
        pvdata_str.append(doc.createTextNode(_a2s(pvdata[:, i])))

    # CELL DATA
    celldata = doc.createElementNS('VTK', 'CellData')
    # cdata
    cdata_obj = []
    cdata_str = []
    for i in range(0, n_cdata):
        cdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type': 'Float32', 'Name': f'cdata {i}',
             'NumberOfComponents': '1', 'format': 'ascii'}
        _set_attributes(d, cdata_obj[i])
        cdata_str.append(doc.createTextNode(_a2s(cdata_all[:, i])))
    # cvdata
    cvdata_obj = []
    cvdata_str = []
    for i in range(0, n_cvdata):
        cvdata_obj.append(doc.createElementNS('VTK', 'DataArray'))
        d = {'type': 'Float32', 'Name': f'cvdata {i}',
             'NumberOfComponents': '3', 'format': 'ascii'}
        _set_attributes(d, cvdata_obj[i])
        cvdata_str.append(doc.createTextNode(_a2s(cvdata_all[:, i])))

    doc.appendChild(root)
    root.appendChild(grid)
    grid.appendChild(piece)

    piece.appendChild(points)
    points.appendChild(points_data)
    points_data.appendChild(points_data_str)

    piece.appendChild(cells)
    cells.appendChild(cells_data)
    cells.appendChild(cells_offset_data)
    cells.appendChild(cells_type_data)
    cells_data.appendChild(cells_data_str)
    cells_offset_data.appendChild(cells_offset_data_str)
    cells_type_data.appendChild(cells_type_data_str)

    piece.appendChild(pointdata)
    for i in range(0, n_pdata):
        pointdata.appendChild(pdata_obj[i])
        pdata_obj[i].appendChild(pdata_str[i])
    for i in range(0, n_pvdata):
        pointdata.appendChild(pvdata_obj[i])
        pvdata_obj[i].appendChild(pvdata_str[i])

    piece.appendChild(celldata)
    for i in range(0, n_cdata):
        celldata.appendChild(cdata_obj[i])
        cdata_obj[i].appendChild(cdata_str[i])
    for i in range(0, n_cvdata):
        celldata.appendChild(cvdata_obj[i])
        cvdata_obj[i].appendChild(cvdata_str[i])

    try:
        with open(fname, 'w', encoding='utf-8') as f:
            doc.writexml(f, newl='\n')
    except OSError as e:
        print(f'.vtu error ({e.errno}): {e.strerror}')


def write_basic_mesh(V, E2V=None, mesh_type='tri',
                     pdata=None, pvdata=None,
                     cdata=None, cvdata=None, fname='output.vtk'):
    """Write mesh file for basic types of elements.

    Parameters
    ----------
    fname : {string}
        file to be written, e.g. 'mymesh.vtu'
    V : {array}
        coordinate array (N x D)
    E2V : {array}
        element index array (Nel x Nelnodes)
    mesh_type : {string}
        type of elements: tri, quad, tet, hex (all 3d)
    pdata : {array}
        scalar data on vertices (N x Nfields)
    pvdata : {array}
        vector data on vertices (3*Nfields x N)
    cdata : {array}
        scalar data on cells (Nfields x Nel)
    cvdata : {array}
        vector data on cells (3*Nfields x Nel)

    Returns
    -------
    writes a .vtu file for use in Paraview

    Notes
    -----
    The difference between write_basic_mesh and write_vtu is that write_vtu is
    more general and requires dictionaries of cell information.
    write_basic_mesh calls write_vtu

    Examples
    --------
    >>> import numpy as np
    >>> from pyamg.vis import write_basic_mesh
    >>> V = np.array([[0.0,0.0],
    ...               [1.0,0.0],
    ...               [2.0,0.0],
    ...               [0.0,1.0],
    ...               [1.0,1.0],
    ...               [2.0,1.0],
    ...               [0.0,2.0],
    ...               [1.0,2.0],
    ...               [2.0,2.0],
    ...               [0.0,3.0],
    ...               [1.0,3.0],
    ...               [2.0,3.0]])
    >>> E2V = np.array([[0,4,3],
    ...                 [0,1,4],
    ...                 [1,5,4],
    ...                 [1,2,5],
    ...                 [3,7,6],
    ...                 [3,4,7],
    ...                 [4,8,7],
    ...                 [4,5,8],
    ...                 [6,10,9],
    ...                 [6,7,10],
    ...                 [7,11,10],
    ...                 [7,8,11]])
    >>> pdata=np.ones((12,2))
    >>> pvdata=np.ones((12*3,2))
    >>> cdata=np.ones((12,2))
    >>> cvdata=np.ones((3*12,2))
    >>> write_basic_mesh(V, E2V=E2V, mesh_type='tri',pdata=pdata,
    ...                  pvdata=pvdata, cdata=cdata, cvdata=cvdata,
    ...                  fname='test.vtu')

    See Also
    --------
    write_vtu

    """
    if E2V is None:
        mesh_type = 'vertex'

    map_type_to_key = {'vertex': 1, 'tri': 5, 'quad': 9, 'tet': 10, 'hex': 12}

    if mesh_type not in map_type_to_key:
        raise ValueError(f'Unknown mesh_type={mesh_type}')

    key = map_type_to_key[mesh_type]

    if mesh_type == 'vertex':
        uidx = np.arange(0, V.shape[0]).reshape((V.shape[0], 1))
        E2V = {key: uidx}
    else:
        E2V = {key: E2V}

    if cdata is not None:
        cdata = {key: cdata}

    if cvdata is not None:
        cvdata = {key: cvdata}

    write_vtu(V=V, cells=E2V, pdata=pdata, pvdata=pvdata,
              cdata=cdata, cvdata=cvdata, fname=fname)


def _set_attributes(d, elm):
    """Set attributes from dictionary of values."""
    for key in d:
        elm.setAttribute(key, d[key])


def _a2s(a):
    """Convert to string."""
    newstr = ''
    return newstr.join([f'{v} ' for v in a.ravel()])



"""Generates simple meshes."""


import numpy as np


def regular_triangle_mesh(nx, ny):
    """Construct a regular triangular mesh in the unit square.

    Parameters
    ----------
    nx : int
       Number of nodes in the x-direction
    ny : int
       Number of nodes in the y-direction

    Returns
    -------
    Vert : array
        nx*ny x 2 vertex list
    E2V : array
        Nex x 3 element list

    Examples
    --------
    >>> from pyamg.gallery import regular_triangle_mesh
    >>> E2V,Vert = regular_triangle_mesh(3, 2)

    """
    nx, ny = int(nx), int(ny)

    if nx < 2 or ny < 2:
        raise ValueError(f'minimum mesh dimension is 2: {(nx, ny)}')

    Vert1 = np.tile(np.arange(0, nx-1), ny - 1) +\
        np.repeat(np.arange(0, nx * (ny - 1), nx), nx - 1)
    Vert3 = np.tile(np.arange(0, nx-1), ny - 1) +\
        np.repeat(np.arange(0, nx * (ny - 1), nx), nx - 1) + nx
    Vert2 = Vert3 + 1
    Vert4 = Vert1 + 1

    Verttmp = np.meshgrid(np.arange(0, nx, dtype='float'),
                          np.arange(0, ny, dtype='float'))
    Verttmp = (Verttmp[0].ravel(), Verttmp[1].ravel())
    Vert = np.vstack(Verttmp).transpose()
    Vert[:, 0] = (1.0 / (nx - 1)) * Vert[:, 0]
    Vert[:, 1] = (1.0 / (ny - 1)) * Vert[:, 1]

    E2V1 = np.vstack((Vert1, Vert2, Vert3)).transpose()
    E2V2 = np.vstack((Vert1, Vert4, Vert2)).transpose()
    E2V = np.vstack((E2V1, E2V2))

    return Vert, E2V



"""Test MultilevelSolver class."""
import numpy as np
from numpy.testing import TestCase, assert_almost_equal, assert_equal
from scipy import sparse

from pyamg.gallery import poisson
from pyamg.multilevel import coarse_grid_solver, MultilevelSolver


def precon_norm(v, ml):
    """Calculate preconditioner norm of v."""
    v = np.ravel(v)
    w = ml.aspreconditioner()*v
    return np.sqrt(np.dot(v.conjugate(), w))


class TestMultilevel(TestCase):
    def test_coarse_grid_solver(self):
        cases = []

        cases.append(sparse.csr_matrix(np.diag(np.arange(1, 5, dtype=float))))
        cases.append(poisson((4,), format='csr'))
        cases.append(poisson((4, 4), format='csr'))

        from pyamg.krylov import cg

        def fn(A, b):
            return cg(A, b)[0]

        # method should be almost exact for small matrices
        for A in cases:
            for solver in ['splu', 'pinv', 'pinv2', 'lu', 'cholesky',
                           'cg', fn]:
                s = coarse_grid_solver(solver)

                b = np.arange(A.shape[0], dtype=A.dtype)

                x = s(A, b)
                assert_almost_equal(A*x, b)

                # subsequent calls use cached data
                x = s(A, b)
                assert_almost_equal(A*x, b)

    def test_aspreconditioner(self):
        from pyamg import smoothed_aggregation_solver
        from scipy.sparse.linalg import cg
        from pyamg.krylov import fgmres
        np.random.seed(1331277597)

        A = poisson((50, 50), format='csr')
        b = np.random.rand(A.shape[0])

        ml = smoothed_aggregation_solver(A)

        for cycle in ['V', 'W', 'F']:
            M = ml.aspreconditioner(cycle=cycle)
            x, info = cg(A, b, tol=1e-8, maxiter=30, M=M, atol='legacy')
            # cg satisfies convergence in the preconditioner norm
            assert precon_norm(b - A*x, ml) < 1e-8*precon_norm(b, ml)

        for cycle in ['AMLI']:
            M = ml.aspreconditioner(cycle=cycle)
            res = []
            x, info = fgmres(A, b, tol=1e-8, maxiter=30, M=M, residuals=res)
            # fgmres satisfies convergence in the 2-norm
            assert np.linalg.norm(b - A*x) < 1e-8*np.linalg.norm(b)

    def test_accel(self):
        from pyamg import smoothed_aggregation_solver
        from pyamg.krylov import cg, bicgstab
        np.random.seed(30459128)

        A = poisson((50, 50), format='csr')
        b = np.random.rand(A.shape[0])

        ml = smoothed_aggregation_solver(A)

        # cg halts based on the preconditioner norm
        for accel in ['cg', cg]:
            x = ml.solve(b, maxiter=30, tol=1e-8, accel=accel)
            assert precon_norm(b - A*x, ml) < 1e-8*precon_norm(b, ml)
            residuals = []
            x = ml.solve(b, maxiter=30, tol=1e-8, residuals=residuals,
                         accel=accel)
            assert precon_norm(b - A*x, ml) < 1e-8*precon_norm(b, ml)
            # print residuals
            assert_almost_equal(precon_norm(b - A*x, ml), residuals[-1])

        # cgs and bicgstab use the Euclidean norm
        for accel in ['bicgstab', 'cgs', bicgstab]:
            x = ml.solve(b, maxiter=30, tol=1e-8, accel=accel)
            assert np.linalg.norm(b - A*x) < 1e-8*np.linalg.norm(b)
            residuals = []
            x = ml.solve(b, maxiter=30, tol=1e-8, residuals=residuals,
                         accel=accel)
            assert np.linalg.norm(b - A*x) < 1e-8*np.linalg.norm(b)
            # print residuals
            assert_almost_equal(np.linalg.norm(b - A*x), residuals[-1])

    def test_cycle_complexity(self):
        # four levels
        levels = []
        levels.append(MultilevelSolver.Level())
        levels[0].A = sparse.csr_matrix(np.ones((10, 10)))
        levels[0].P = sparse.csr_matrix(np.ones((10, 5)))
        levels.append(MultilevelSolver.Level())
        levels[1].A = sparse.csr_matrix(np.ones((5, 5)))
        levels[1].P = sparse.csr_matrix(np.ones((5, 3)))
        levels.append(MultilevelSolver.Level())
        levels[2].A = sparse.csr_matrix(np.ones((3, 3)))
        levels[2].P = sparse.csr_matrix(np.ones((3, 2)))
        levels.append(MultilevelSolver.Level())
        levels[3].A = sparse.csr_matrix(np.ones((2, 2)))

        # one level hierarchy
        mg = MultilevelSolver(levels[:1])
        assert_equal(mg.cycle_complexity(cycle='V'), 100.0/100.0)  # 1
        assert_equal(mg.cycle_complexity(cycle='W'), 100.0/100.0)  # 1
        assert_equal(mg.cycle_complexity(cycle='AMLI'), 100.0/100.0)  # 1
        assert_equal(mg.cycle_complexity(cycle='F'), 100.0/100.0)  # 1

        # two level hierarchy
        mg = MultilevelSolver(levels[:2])
        assert_equal(mg.cycle_complexity(cycle='V'), 225.0/100.0)  # 2,1
        assert_equal(mg.cycle_complexity(cycle='W'), 225.0/100.0)  # 2,1
        assert_equal(mg.cycle_complexity(cycle='AMLI'), 225.0/100.0)  # 2,1
        assert_equal(mg.cycle_complexity(cycle='F'), 225.0/100.0)  # 2,1

        # three level hierarchy
        mg = MultilevelSolver(levels[:3])
        assert_equal(mg.cycle_complexity(cycle='V'), 259.0/100.0)  # 2,2,1
        assert_equal(mg.cycle_complexity(cycle='W'), 318.0/100.0)  # 2,4,2
        assert_equal(mg.cycle_complexity(cycle='AMLI'), 318.0/100.0)  # 2,4,2
        assert_equal(mg.cycle_complexity(cycle='F'), 318.0/100.0)  # 2,4,2

        # four level hierarchy
        mg = MultilevelSolver(levels[:4])
        assert_equal(mg.cycle_complexity(cycle='V'), 272.0/100.0)  # 2,2,2,1
        assert_equal(mg.cycle_complexity(cycle='W'), 388.0/100.0)  # 2,4,8,4
        assert_equal(mg.cycle_complexity(cycle='AMLI'), 388.0/100.0)  # 2,4,8,4
        assert_equal(mg.cycle_complexity(cycle='F'), 366.0/100.0)  # 2,4,6,3


class TestComplexMultilevel(TestCase):
    def test_coarse_grid_solver(self):
        cases = []

        cases.append(sparse.csr_matrix(np.diag(np.arange(1, 5))))
        cases.append(poisson((4,), format='csr'))
        cases.append(poisson((4, 4), format='csr'))

        # Make cases complex
        cases = [G+1e-5j*G for G in cases]
        cases = [0.5*(G + G.H) for G in cases]

        # method should be almost exact for small matrices
        for A in cases:
            for solver in ['splu', 'pinv', 'pinv2', 'lu', 'cholesky', 'cg']:
                s = coarse_grid_solver(solver)

                b = np.arange(A.shape[0], dtype=A.dtype)

                x = s(A, b)
                assert_almost_equal(A*x, b)

                # subsequent calls use cached data
                x = s(A, b)
                assert_almost_equal(A*x, b)



"""Relaxation-centric utilities."""

import numpy as np
from scipy.sparse.linalg import LinearOperator
from ..multilevel import MultilevelSolver
from .. import relaxation


def relaxation_as_linear_operator(method, A, b):
    """Create a linear operator that applies a relaxation method to a right-hand-side.

    Parameters
    ----------
    methods : {tuple or string}
        Relaxation descriptor: Each tuple must be of the form ('method','opts')
        where 'method' is the name of a supported smoother, e.g., gauss_seidel,
        and 'opts' a dict of keyword arguments to the smoother, e.g., opts =
        {'sweep':symmetric}.  If string, must be that of a supported smoother,
        e.g., gauss_seidel.

    Returns
    -------
    linear operator that applies the relaxation method to a vector for a
    fixed right-hand-side, b.

    Notes
    -----
    This method is primarily used to improve B during the aggregation setup
    phase.  Here b = 0, and each relaxation call can improve the quality of B,
    especially near the boundaries.

    Examples
    --------
    >>> from pyamg.gallery import poisson
    >>> from pyamg.relaxation.utils import relaxation_as_linear_operator
    >>> import numpy as np
    >>> A = poisson((100,100), format='csr')           # matrix
    >>> B = np.ones((A.shape[0],1))                 # Candidate vector
    >>> b = np.zeros((A.shape[0]))                  # RHS
    >>> relax = relaxation_as_linear_operator('gauss_seidel', A, b)
    >>> B = relax*B

    """

    def unpack_arg(v):
        if isinstance(v, tuple):
            return v[0], v[1]
        return v, {}

    # setup variables
    accepted_methods = ['gauss_seidel', 'block_gauss_seidel', 'sor',
                        'gauss_seidel_ne', 'gauss_seidel_nr', 'jacobi',
                        'block_jacobi', 'richardson', 'schwarz',
                        'strength_based_schwarz', 'jacobi_ne']

    b = np.array(b, dtype=A.dtype)
    fn, kwargs = unpack_arg(method)
    lvl = MultilevelSolver.Level()
    lvl.A = A

    # Retrieve setup call from relaxation.smoothing for this relaxation method
    if fn not in accepted_methods:
        raise NameError(f'invalid relaxation method: {fn}')
    try:
        setup_smoother = getattr(relaxation.smoothing, 'setup_' + fn)
    except NameError as e:
        raise NameError(f'invalid presmoother method: {fn}') from e

    # Get relaxation routine that takes only (A, x, b) as parameters
    relax = setup_smoother(lvl, **kwargs)

    # Define matvec
    def matvec(x):
        xcopy = x.copy()
        relax(A, xcopy, b)
        return xcopy

    return LinearOperator(A.shape, matvec, dtype=A.dtype)



"""Generalized Minimum Residual Method (GMRES) Krylov solver."""

from ._gmres_mgs import gmres_mgs
from ._gmres_householder import gmres_householder


def gmres(A, b, x0=None, tol=1e-5, restrt=None, maxiter=None,
          M=None, callback=None, residuals=None, orthog='householder',
          **kwargs):
    """Generalized Minimum Residual Method (GMRES).

    GMRES iteratively refines the initial solution guess to the
    system Ax = b.  Left preconditioned.  Residuals are preconditioned residuals.

    Parameters
    ----------
    A : array, matrix, sparse matrix, LinearOperator
        n x n, linear system to solve
    b : array, matrix
        right hand side, shape is (n,) or (n,1)
    x0 : array, matrix
        initial guess, default is a vector of zeros
    tol : float
        Tolerance for stopping criteria, let r=r_k
        ||M r|| < tol ||M b||
        if ||b||=0, then set ||M b||=1 for these tests.
    restrt : None, int
        - if int, restrt is max number of inner iterations
          and maxiter is the max number of outer iterations
        - if None, do not restart GMRES, and max number of inner iterations
          is maxiter
    maxiter : None, int
        - if restrt is None, maxiter is the max number of inner iterations
          and GMRES does not restart
        - if restrt is int, maxiter is the max number of outer iterations,
          and restrt is the max number of inner iterations
        - defaults to min(n,40) if restart=None
    M : array, matrix, sparse matrix, LinearOperator
        n x n, inverted preconditioner, i.e. solve M A x = M b.
    callback : function
        User-supplied function is called after each iteration as
        callback(xk), where xk is the current solution vector
    residuals : list
        preconditioned residual history in the 2-norm, including the initial residual
    orthog : string
        'householder' calls _gmres_householder which uses Householder
        reflections to find the orthogonal basis for the Krylov space.
        'mgs' calls _gmres_mgs which uses modified Gram-Schmidt to find the
        orthogonal basis for the Krylov space

    Returns
    -------
    (xk, info)
    xk : an updated guess after k iterations to the solution of Ax = b
    info : halting status

            ==  =======================================
            0   successful exit
            >0  convergence to tolerance not achieved,
                return iteration count instead.
            <0  numerical breakdown, or illegal input
            ==  =======================================

    Notes
    -----
    The LinearOperator class is in scipy.sparse.linalg.
    Use this class if you prefer to define A or M as a mat-vec routine
    as opposed to explicitly constructing the matrix.

    The orthogonalization method, orthog='householder', is more robust
    than orthog='mgs', however for the majority of problems your
    problem will converge before 'mgs' loses orthogonality in your basis.

    orthog='householder' has been more rigorously tested, and is
    therefore currently the default

    The residual is the *preconditioned* residual.


    Examples
    --------
    >>> from pyamg.krylov import gmres
    >>> from pyamg.util.linalg import norm
    >>> import numpy as np
    >>> from pyamg.gallery import poisson
    >>> A = poisson((10,10))
    >>> b = np.ones((A.shape[0],))
    >>> (x,flag) = gmres(A,b, maxiter=2, tol=1e-8)
    >>> print(f'{norm(b - A*x):.6}')
    6.54282

    References
    ----------
    .. [1] Yousef Saad, "Iterative Methods for Sparse Linear Systems,
       Second Edition", SIAM, pp. 151-172, pp. 272-275, 2003
       http://www-users.cs.umn.edu/~saad/books.html

    """
    # pass along **kwargs
    if orthog == 'householder':
        (x, flag) = gmres_householder(A, b, x0=x0, tol=tol, restrt=restrt,
                                      maxiter=maxiter, M=M,
                                      callback=callback, residuals=residuals,
                                      **kwargs)
    elif orthog == 'mgs':
        (x, flag) = gmres_mgs(A, b, x0=x0, tol=tol, restrt=restrt,
                              maxiter=maxiter, M=M,
                              callback=callback, residuals=residuals, **kwargs)

    return (x, flag)



"""Krylov solvers.

This module contains several Krylov subspace methods, in addition to two simple
iterations, to solve linear systems iteratively.  These methods often use
multigrid as a preconditioner to accelerate convergence to the solution.  See [1]_ and [2]_.

Functions
---------
    - gmres
    - fgmres
    - cgne
    - cgnr
    - cg
    - bicgstab
    - steepest descent, (simple iteration)
    - minimal residual (MR), (simple iteration)


References
----------
.. [1] Yousef Saad, "Iterative Methods for Sparse Linear Systems,
   Second Edition", SIAM, pp. 231-234, 2003
   http://www-users.cs.umn.edu/~saad/books.html

.. [2] Richard Barrett et al.  "Templates for the Solution of Linear Systems:
   Building Blocks for Iterative Methods, 2nd Edition", SIAM
   http://www.netlib.org/linalg/html_templates/Templates.html
   http://www.netlib.org/templates/

"""

from ._gmres import gmres
from ._gmres_householder import gmres_householder
from ._gmres_mgs import gmres_mgs
from ._fgmres import fgmres
from ._cg import cg
from ._cr import cr
from ._cgnr import cgnr
from ._cgne import cgne
from ._bicgstab import bicgstab
from ._steepest_descent import steepest_descent
from ._minimal_residual import minimal_residual

__all__ = [
    'gmres',
    'gmres_householder',
    'gmres_mgs',
    'fgmres',
    'cg',
    'cr',
    'cgnr',
    'cgne',
    'bicgstab',
    'steepest_descent',
    'minimal_residual'
]



"""Test simple iteration."""
from functools import partial
import numpy as np
from numpy.testing import TestCase

import pyamg
from pyamg.krylov import minimal_residual, steepest_descent
from pyamg.util.linalg import norm


class TestSimpleIterations(TestCase):
    def setUp(self):
        self.definite_cases = []
        self.spd_cases = []

        # 1x1
        A = np.array([[1.2]])
        b = np.array([3.9]).reshape(-1, 1)
        x0 = np.zeros((1, 1))
        self.definite_cases.append({'A': A, 'b': b, 'x0': x0, 'maxiter': 1,
                                    'reduction_factor': 1e-10})
        self.spd_cases.append({'A': A, 'b': b, 'x0': x0, 'maxiter': 1,
                               'reduction_factor': 1e-10})

        # 2x2
        A = np.array([[4.5, -1.2],
                      [3.4, 6.7]])
        b = np.array([-3.2, 5.6]).reshape(-1, 1)
        x0 = np.zeros((2, 1))
        self.definite_cases.append({'A': A, 'b': b, 'x0': x0, 'maxiter': 4,
                                    'reduction_factor': 0.1})
        self.spd_cases.append({'A': A.T.dot(A), 'b': b, 'x0': x0, 'maxiter': 4,
                               'reduction_factor': 0.1})

        # 2x2 Imaginary
        A = np.array([[4.5, -1.2j],
                      [3.4+1.2j, 6.7]])
        b = np.array([-3.2, 5.6]).reshape(-1, 1)
        x0 = np.zeros((2, 1))
        self.spd_cases.append({'A': A.T.dot(A), 'b': b, 'x0': x0, 'maxiter': 4,
                               'reduction_factor': 0.1})

        # 4x4
        A = np.array([[1.2, 0., 0., 0.],
                      [0., 10., 2., 6.],
                      [0., 0., 9.3, -2.31],
                      [-4., 0., 0., 11.]])
        b = np.array([1., 3.9, 0., -1.23]).reshape(-1, 1)
        x0 = np.zeros((4, 1))
        self.definite_cases.append({'A': A, 'b': b, 'x0': x0, 'maxiter': 4,
                                    'reduction_factor': 0.3})
        self.spd_cases.append({'A': A.T.dot(A), 'b': b, 'x0': x0, 'maxiter': 4,
                               'reduction_factor': 0.3})

        # 4x4 Imaginary
        A = np.array(A, dtype=complex)
        A[0, 0] += 3.1j
        A[3, 3] -= 1.34j
        A[1, 3] *= 1.0j
        A[1, 2] += 1.0j
        b = np.array([1. - 1.0j, 2.0 - 3.9j, 0., -1.23]).reshape(-1, 1)
        x0 = np.ones((4, 1))
        self.spd_cases.append({'A': A.conj().T.dot(A), 'b': b, 'x0': x0, 'maxiter': 4,
                               'reduction_factor': 0.3})

        # 10x10
        A = np.array([[91.1, 0., 0., 0., 3.9, 0., 0., 11., -1., 0.],
                      [0., 45., 2.9, 0., 0., 6.8, 0., 0., 0., 0.],
                      [0., 0., 19.0, 0., 0., 0.8, 1., -2.2, 0., 9.],
                      [-4., 0., 0.0, 4., 0., 0.0, 2., 2.2, 0., 0.],
                      [0., 0., 0.0, 21., 30., 0.1, 0., 0., 0., 0.],
                      [0., 0., 0.0, 0., -4.7, 7.0, 0., 0., 0., 0.],
                      [2.1, 7., 2.0, 0., 0., 0.0, 8., 0., 0., 0.],
                      [0., 0., 0.0, 34., 0., 0.0, 0., 87.1, -12.3, 0.],
                      [0., 3.4, 0.0, 0., 0., -0.3, 0., 0., 7., 0.],
                      [9., 0., 0.0, 0., 8.7, 0.0, 0., 0., 0., 11.2]])
        b = np.array([1., 0., 0.2, 8., 0., -1.9,
                      11.3, 0.0, 0.1, 0.0]).reshape(-1, 1)
        x0 = np.zeros((10, 1))
        x0[4] = 11.1
        x0[7] = -2.1
        self.definite_cases.append({'A': A, 'b': b, 'x0': x0, 'maxiter': 2,
                                    'reduction_factor': 0.98})
        self.spd_cases.append({'A':
                               pyamg.gallery.poisson((10,)).toarray(),
                               'b': b, 'x0': x0, 'maxiter': 2,
                               'reduction_factor': 0.5})

    def test_steepest_descent(self):
        # Ensure repeatability
        np.random.seed(0)

        def cb(x, A, b):
            fvals.append(0.5*np.dot(x.ravel(), np.ravel(A @ x))
                         - np.dot(b.ravel(), x.ravel()))

        for case in self.spd_cases:
            A = case['A']
            b = case['b']
            x0 = case['x0']
            maxiter = case['maxiter']
            reduction_factor = case['reduction_factor']

            # This function should always decrease
            fvals = []

            callback = partial(cb, A=A, b=b)

            x, _ = steepest_descent(A, b, x0=x0, tol=1e-16,
                                    maxiter=maxiter, callback=callback)
            norm1 = norm(np.ravel(b) - np.ravel(A.dot(x.reshape(-1, 1))))
            norm2 = norm(np.ravel(b) - np.ravel(A.dot(x0.reshape(-1, 1))))
            actual_factor = norm1 / norm2
            assert actual_factor < reduction_factor

            if A.dtype != complex:
                for i in range(len(fvals)-1):
                    assert fvals[i+1] <= fvals[i]

        # Test preconditioning
        A = pyamg.gallery.poisson((10, 10), format='csr')
        b = np.random.rand(A.shape[0], 1)
        x0 = np.random.rand(A.shape[0], 1)
        fvals = []

        callback = partial(cb, A=A, b=b)

        resvec = []
        sa = pyamg.smoothed_aggregation_solver(A)
        x, _ = steepest_descent(A, b, x0, tol=1e-8, maxiter=20,
                                residuals=resvec, M=sa.aspreconditioner(),
                                callback=callback)
        assert resvec[-1]/resvec[0] < 1e-8
        for i in range(len(fvals)-1):
            assert fvals[i+1] <= fvals[i]

    def test_minimal_residual(self):
        # Ensure repeatability
        np.random.seed(0)

        self.definite_cases.extend(self.spd_cases)

        def cb(x, A):
            fvals.append(np.sqrt(np.dot(x.ravel(), np.ravel(A @ x))))

        for case in self.definite_cases:
            A = case['A']
            maxiter = case['maxiter']
            x0 = np.random.rand(A.shape[0],)
            b = np.zeros_like(x0)
            reduction_factor = case['reduction_factor']
            if A.dtype != complex:

                # This function should always decrease (assuming zero RHS)
                fvals = []

                callback = partial(cb, A=A)
                #
                x, _ = minimal_residual(A, b, x0=x0,
                                        tol=1e-16, maxiter=maxiter,
                                        callback=callback)
                norm1 = norm(np.ravel(b) - np.ravel(A.dot(x.reshape(-1, 1))))
                norm2 = norm(np.ravel(b) - np.ravel(A.dot(x0.reshape(-1, 1))))
                actual_factor = norm1 / norm2
                assert actual_factor < reduction_factor
                if A.dtype != complex:
                    for i in range(len(fvals)-1):
                        assert fvals[i+1] <= fvals[i]

        # Test preconditioning
        A = pyamg.gallery.poisson((10, 10), format='csr')
        x0 = np.random.rand(A.shape[0], 1)
        b = np.zeros_like(x0)
        fvals = []

        callback = partial(cb, A=A)
        #
        resvec = []
        sa = pyamg.smoothed_aggregation_solver(A)
        x, _ = minimal_residual(A, b, x0, tol=1e-8, maxiter=20,
                                residuals=resvec, M=sa.aspreconditioner(),
                                callback=callback)
        assert resvec[-1]/resvec[0] < 1e-8
        for i in range(len(fvals)-1):
            assert fvals[i+1] <= fvals[i]
